#ifndef HISTORY_RECORD_H
#define HISTORY_RECORD_H
//HistoryRecord.h
/*
  (c)2023 Forward Computing and Control Pty. Ltd.
  NSW Australia, www.forward.com.au
  This code is not warranted to be fit for any purpose. You may only use it at your own risk.
  This code may be freely used for both private and commercial use
  Provided this copyright is maintained.
*/

#include <time.h>

// Structure for storing binary records with timestamp
struct BinaryRecordStruct {
  time_t utc_time;          // UTC time in seconds since epoch
  uint32_t utc_usec;          // UTC time usec part
  int32_t power_w;          // Power in Watts
  int32_t hw_w;             // Hot Water in Watts
  double hw_kwh;            // Hot Water in kWh
//  double offpeak_kwh;       // Off-peak in kWh
  uint8_t state;            // 0=Off, 1=Solar Hot Water, 2=Off Peak
  uint8_t hw_triac;         // 0=off, 1=on
  uint8_t hw_boost;         // 0=off, 1=on
};

struct  Accumulated_kWh_RecordStruct {
  double export_kWhrs;
  double import_kWhrs;
  double hotwater_kWhrs;
  double offPeak_kWhrs;
  double cost; // for today sofar
  double lastPower_kW; // for average
  double lastHW_kW; // for average
  uint64_t lasttime_ms;
  int64_t t_offset; 
  bool fullyHeated;  //<< need this one
  bool lowPowerTracking;
  uint64_t lowPowerStartTime;
};

struct  HistoryRecordStruct {
  // calculate these from todays data
  double export_kWhrs;
  double import_kWhrs;
  double hotwater_kWhrs;
  double offPeak_kWhrs;
  double cost; // for today sofar
  bool fullyHeated;  //<< need this one
  double export_kWhrs_yesterday; // << calculated
  double import_kWhrs_yesterday; // << calculated
  double hotwater_kWhrs_yesterday; // << calculated
  double offPeak_kWhrs_yesterday; // << calculated
  double cost_yesterday; // for today sofar
  bool fullyHeated_yesterday;
};

#endif
