// SolarHotWater Monitor V2 (no controls)
// programmed on Fire_Beetle ESP32 E N16R2 (16M flash / 2M SPRAM with ESP32 board support V3.0.7
// has user switch connected to D4

/***************
 NOTE: This sketch needs a modified boards.txt file and an additional partition file
 see https://www.forward.com.au/pfod/HomeAutomation/SolarHotWater/Ver2/index.html
 
 the board.txt file needs these lines
dfrobot_firebeetle2_esp32e.upload.maximum_size=2097152  << modify existing line
// add these line
dfrobot_firebeetle2_esp32e.menu.PartitionScheme.huge_spiffs=16M with spiffs (2MB APP/13.9MB SPIFFS)
dfrobot_firebeetle2_esp32e.menu.PartitionScheme.huge_spiffs.build.partitions=app2MB_spiffs_rest

The additional partition file is app2MB_spiffs_rest.csv
# Name,    Type,    SubType,    Offset,    Size,    Flags
nvs,       data,    nvs,        0x9000,    0x5000,    
otadata,   data,    ota,        0xe000,    0x2000,    
app0,      app,     ota_0,      0x10000,   0x200000,    
spiffs,    data,    spiffs,     0x210000,  0xde0000,    
coredump,  data,    coredump,   0xff0000,  0x10000,    

***********/

// Using ESP32 based board programmed via Arduino IDE
// follow the steps given on http://www.forward.com.au/pfod/ESP32/index.html to install ESP32 support for Arduino IDE

/* Based on Code generated by pfodDesignerV3 V3.0.4181
*/
/*
   (c)2023 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This generated code may be freely used for both private and commercial use
   provided this copyright is maintained.
*/

#define CONFIG_FREERTOS_NUMBER_OF_CORES 2
#include "esp_task_wdt.h"

#include "RTC_NTP_replacement.h"
#include <pfodEEPROM.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include "millisDelay.h"
#include "SafeString.h"
#include "SafeStringReader.h"
#include "BufferedOutput.h"
#include <pfodDwgs.h>
#include "Dwg_1.h"
#include "Label.h"
#include <ESPAutoWiFiConfig.h>
#include "processData.h"
#include "volatileVars.h"
#include "HW_webpages.h"
#include "SafeStringReader.h"
#include "HistoryFile.h"
#include "loopTimer.h"
#include "DebouncedSwitch.h"
#include "locks.h"
#include "systemTimeUtils.h" // some time functions
#include "csvProcessor.h" // CSV file processing

const int SQUARE_WAVE_PIN = 15; // the pin the RTC 1Hz output is connected to

//#define DEBUG
//#define STARTUP_DELAY
// delay start by 5 sec

//#define CLIENT_DEBUG
// this next define is used for testing. Suppresses connection to hot water controller so as not to interfer with currently running HW monitor.
//#define NO_TELNET_CONNECT

// enables loop timer on processing hot water telnet connection
//#define HOT_WATER_CLIENT_TIME_CHECK

static Print* debugPtr = NULL;

createBufferedOutput(output, 512, DROP_IF_FULL); // buffered out with variable name output with buffer size 512 and mode drop chars if buffer full

void startTimers(unsigned int excessSec); // day and reboot


/**
  Set a compiled default TZ here. Can be overrided/edited later by webpage.
  get_DefaultTZ() weakly defined in TZ_support.cpp
  if const char *get_DefaultTZ() is not defined then GMT0 is used as the default.
  See
  https://github.com/nayarsystems/posix_tz_db/blob/master/zones.csv
  for a list of time zones and their posix_tz strings
  that can be copied and pasted into the get_DefaultTZ or into the Set Time Zone webpage.
*/
const char *get_DefaultTZ() { // magic name picked up by TZ_Support.cpp
  static char _default_tz_[50] = "AEST-10AEDT,M10.1.0,M4.1.0/3";   // TZ_Australia_Sydney
  return _default_tz_;
}

static millisDelay restartDelay;
const static unsigned long RESTART_DELAY_ms = 2ul * 60 * 1000; // reset after 2min

static millisDelay hrCheckTimer; // check time for roll history at midnight and reboot at 4am
static bool timersStarted = false;

static void handleHotWaterClient();
void handle_pfodServerConnection();
createSafeStringReader(sfCmdReader, 80, "\r\n"); // for commands

DebouncedSwitch sw(D4); // switch for force save of data

int ledPin = -D8; // for Sparkfun ESP32 IoT Redboard // D7; // for XIAO esp32c3 external led // -D8 for Fire_Beetle, -8; for ESP32-C3
bool highForLedOn = true; //
size_t eepromOffset = 32; // allow for security if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data

const char* HotWater_host = "10.1.1.119"; // HotWater controller and power monitor
int HotWater_portNo = 23; // telnet


millisDelay failedConnectRebootDelay;
unsigned long FAILED_CONNECT_REBOOT_DELAY_MS = 1ul * 60 * 1000; // reboot in 1min

// Monitor No control from this app

//const unsigned MAX_HotWaterOffCount = 5; // 25sec delay before marking hotwater fully charged

int status;     // the Wifi radio's status

#include <WiFi.h>

// Download pfodESP32BufferedClient library from http://www.forward.com.au/pfod/pfodParserLibraries/index.html
// pfodESP32BufferedClient.zip contains pfodESP32BufferedClient and pfodESP32Utils
#include <pfodESP32Utils.h>
#include <pfodESP32BufferedClient.h>

// download the libraries from http://www.forward.com.au/pfod/pfodParserLibraries/index.html


#include <pfodSecurity.h>
int swap01(int); // method prototype for slider end swaps
char version[] = "V3";
pfodParser parser(version); // always have this one // create a parser with menu version string to handle the pfod messages
pfodDwgs dwgs(&parser);  // drawing support
void sendMainMenu(pfodParser & parser);
void sendMainMenuUpdate(pfodParser & parser);
void closeConnection(Stream * io);
pfodESP32BufferedClient bufferedClient;


const uint8_t MAX_CLIENTS = 4; // the default
WiFiClient clients[MAX_CLIENTS]; // hold the currently open clients
pfodParser *parserPtrs[MAX_CLIENTS]; // hold the parsers for each client
pfodESP32BufferedClient *bufferedClientPtrs[MAX_CLIENTS]; // hold the parsers for each client

const int portNo = 4989; // What TCP port to listen on for connections.

// sfTime input 00:00 to 23:59, output 12:00am to 11:59pm
// input returned unchanged on errors
void convertDayMinsToAmPm(SafeString &sfTime, unsigned long dayMins) {
  int hrs = dayMins / 60;
  int mins = dayMins - hrs * 60;

  bool isPM = false;
  if (hrs >= 12) {
    hrs -= 12;
    isPM = true;
  }
  if (hrs == 0) {
    hrs += 12; // 00:.. => 12:.. am
  }
  sfTime.clear();
  sfTime.print(hrs);
  sfTime.print(":");
  if (mins < 10) {
    sfTime.print('0');
  }
  sfTime.print(mins);
  if (isPM) {
    sfTime += "pm";
  } else {
    sfTime += "am";
  }
}


// sfTime input 00:00 to 23:59, output 12:00am to 11:59pm
// input returned unchanged on errors
void convert24hrToAmPm(SafeString &sfTime) {
  cSF(sfStr, 10);
  cSF(sfMins, 10);
  if (sfTime.capacity() < 7) {
    return; // to small
  }
  if (sfTime.length() != 5) {
    return;
  }
  int idx = sfTime.indexOf(':');
  if (idx != 2) {
    return;
  }
  sfTime.substring(sfStr, 0, idx);
  sfTime.substring(sfMins, idx);
  bool isPM = false;
  int hrs = 0;
  if (sfStr.toInt(hrs)) {
    if (hrs >= 12) {
      hrs -= 12;
      isPM = true;
    }
    if (hrs == 0) {
      hrs += 12; // 00:.. => 12:.. am
    }
    sfTime.clear();
    sfTime.print(hrs);
    sfTime += sfMins;
    if (isPM) {
      sfTime += "pm";
    } else {
      sfTime += "am";
    }
  } else {
    return; // input unchanged
  }
}



// The implementation of CSV file processing has been moved to csvProcessor.cpp

pfodAutoCmd mainDwgLoadCmd; // the main empty dwg
static bool forceMainMenuReload = true; // for testing force complete reload on every reboot
static bool forceMainDwgReload = true; // for testing force complete reload on every reboot
static unsigned long dwgRefresh = 2500;//set to 0 to disable auto refresh or 10000; to  refresh every 10sec
// when on the front screen, the mobile back button will also refresh, otherwise it goes back to the previous menu/screen

Dwg_1 _dwg_1(&parser, &dwgs); // create dwg object and add to parser processing
// Monitor No control from this app
Label hotwaterkWhrs(&parser, &dwgs); // solar + offpeak totals
Label hotwaterFull(&parser, &dwgs); // hot water either solar or boost
Label totalCost(&parser, &dwgs); //
Label yesterday(&parser, &dwgs); //
Label nameTimeDisplay(&parser, &dwgs);
cSF(sfTime, 10); // 00:00 to 23:59 => 12:00am to 12:59pm

// send main dwg that just inserts the designed dwg at the pushZero position and scaling.
bool sendMainDwg(pfodParser & parser) {
  dwgs.setOut(&parser);
  // main dwg just inserts desgined dwg
  dwgs.start(50, 80, dwgs.WHITE, false); //false means NO more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); //need to set a version number for image refresh to work!! send the parser version to cache this image, refreshes every 10sec
  dwgs.pushZero(25, 22, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(_dwg_1).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 47, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(totalCost).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 53, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterkWhrs).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 58, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterFull).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 65, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(yesterday).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 74, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(nameTimeDisplay).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.end();
  return true;
}

void sendMainDwgUpdate(pfodParser & parser) { // nothing here, but insertedDwgs updated by parser.parse()
  dwgs.setOut(&parser);
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  dwgs.end();
}

void hard_restart() {
  appendBufferToFile(); // always

  esp_task_wdt_config_t twdt_config = {
    .timeout_ms = 1000,
    .idle_core_mask = (1 << 2) - 1,    // Bitmask of all cores 2 cores on ESP32 E
    .trigger_panic = true,
  };
  esp_task_wdt_init(&twdt_config);
  esp_task_wdt_add(NULL);
  while (true); // loop here until timeout force reboot.
}



WiFiServer server(portNo);
WiFiClient client;
WiFiClient* clientPtr; // non-null if have connected client

unsigned long plot_msOffset = 0; // set by {@} response
bool clearPlot = false; // set by the {@} response code


void handle_pfodParser(pfodParser & parser) {
  uint8_t cmd = parser.parse(); // parse incoming data from connection
  // parser returns non-zero when a pfod command is fully parsed and not consumed by an inserted dwg
  if (cmd != 0) { // have parsed a complete msg { to }
    if ('.' == cmd) {
      // pfodApp has connected and sent {.} , it is asking for the main menu
      if (parser.isRefresh() && !forceMainMenuReload) {
        sendMainMenuUpdate(parser); // menu is cached just send update
      } else {
        forceMainMenuReload = false;
        sendMainMenu(parser); // send back the menu designed
      }

    } else if (parser.cmdEquals(mainDwgLoadCmd)) {
      if (parser.isRefresh() && !forceMainDwgReload) { // refresh and not forceReload just send update
        sendMainDwgUpdate(parser);
      } else {
        forceMainDwgReload = false; //
        sendMainDwg(parser); // just inserts dwgs and that will update them
      }

    } else if (parser.cmdEquals('A')) { // click in A menu item that holds the main dwg
      // not used as inserted dwgs handle all the clicks
      sendMainMenuUpdate(parser); // always reply to the msg.

      // handle {@} request
    } else if ('@' == cmd) { // pfodApp requested 'current' time
      parser.print(F("{@`0}")); // return `0 as 'current' raw data milliseconds
    } else if ('!' == cmd) {
      // CloseConnection command
      closeConnection(parser.getPfodAppStream());
    } else {
      // unknown command
      parser.print(F("{}")); // always send back a pfod msg otherwise pfodApp will disconnect.
    }
  }
}

// look for dump filename

bool processCmd(SafeString & sfLine) {
  sfLine.trim();
  cSF(sfToken, 50);
  if (sfLine.firstToken(sfToken, ' ')) {
    sfToken.toLowerCase();
    sfToken.trim();
    if (sfToken == "dump") {
      Serial.print("// dump ");
      // look for file name
      if (sfLine.nextToken(sfToken, ' ')) {
        sfToken.trim();
        if ((sfToken == getCostsFilename()) || (sfToken == "costs") || (sfToken == "cost")) {
          Serial.println(getCostsFilename());
          dumpCostsFile(&Serial);
        } else {
          Serial.println(sfToken);
          // try to dump data file
          appendBufferToFile(); // incase we are dumping todays file
          if (!printDataFile(sfToken.c_str(), &Serial)) {
            Serial.println("// unrecognized file name");
            return false;
          }
        }
      } else {
        Serial.println();
        Serial.println("// missing file name");
        return false;
      }
    } else if (sfToken == "load") {
      Serial.println(sfToken);
      Serial.println(" Start loading csv data file.");
      Serial.println(" Line with //FileName:...  must precede first non-comment");
      Serial.println(" other lines starting with // are ignored as are empty lines");
      Serial.println(" other lines starting with // are ignored");
      Serial.println(" OR read times out after 4sec of no input");
      Serial.println(" NOTE: Use Sequential read");
      appendBufferToFile(); // incase we are reading in todays file
      // clears input stream first
      if (readDataStream(Serial, &Serial) ) {
        Serial.println(" Load finished");
        return true;
      } else {
        Serial.println(" Load failed");
        return false;
      }
    } else {
      Serial.print("// Unrecognized cmd: "); Serial.print(sfToken); Serial.print(" "); Serial.println(sfLine);
      return false;
    }
  }
  return true; // empty line
}

// excessSec are left over from last hr
// it timer fires befor the hr
void startTimers(unsigned int excessSec) {
  if (timersStarted) {
    return;
  }
  if (!isRTCDateTimeSet()) { // wait for RTC be set
    return;
  }
  if (debugPtr) {
    debugPtr->print(" startTimers "); debugPtr->print(" excessSec: "); debugPtr->print(excessSec);
  }
  // else find time to next hr local
  unsigned long secLeftInHr = getLocalSecLeftInHr();
  unsigned long msLeft = (secLeftInHr + excessSec) * 1000;
  if (msLeft == 0) {
    msLeft = 1;
  }
  if (debugPtr) {
    debugPtr->print(" check in: "); debugPtr->print(msLeft / 1000.0); debugPtr->println(" secs");
  }
  hrCheckTimer.start(msLeft);
  timersStarted = true;
}



void setup() {
  // Explicitly set the ESP32 to be a WiFi-client, otherwise, it by default,
  // would try to act as both a client and an access-point and could cause
  // network-issues with your other WiFi-devices on your WiFi-network.
  //  WiFi.mode(WIFI_STA);  Fire_Beetle does not like this!!
  EEPROM.begin(512);  // only use 20bytes for pfodSecurity but reserve 512 (pfodWifiConfig uses more)

  Serial.begin(115200); // set Serial for error msgs
  output.connect(Serial); // connect the buffered output
#ifdef DEBUG
#ifdef STARTUP_DELAY
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  Serial.println();
#endif

  debugPtr = &output; // otherwise debugPtr == NULL
#endif

  setDebugOut(debugPtr); // set the debug output for the library, still need to uncomment #define DEBUG in library files

  sfCmdReader.connect(Serial);

  initHistory(); // turn on debug if #define DEBUG in history.cpp

  // always have at least one
  parserPtrs[0] = &parser;
  bufferedClientPtrs[0] = &bufferedClient;

  // fill in the rest
  for (size_t i = 1; i < MAX_CLIENTS; i++) {
    parserPtrs[i] = new pfodParser(version);
    bufferedClientPtrs[i] = new pfodESP32BufferedClient();
  }

  // set initial display state
  hotwaterkWhrs.setTextColor(0);
  hotwaterkWhrs.setTextSize(4);
  hotwaterFull.setTextSize(3);
  hotwaterFull.setTextColor(2);
  totalCost.setTextSize(6);
  yesterday.setTextSize(1);
  nameTimeDisplay.setTextSize(4);
  nameTimeDisplay.setTextColor(12);

#ifdef DEBUG
  // setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, 0, "HotWaterMonitor")) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }

#ifdef DEBUG
  Serial.println();
  Serial.println("WiFi connected");
#endif
  wifiConnected_v = true;

  initRTC(SDA, SCL, SQUARE_WAVE_PIN ); // has 2sec startup delay
  if (!isRTCDateTimeSet()) {
    Serial.println("RTC needs date set");
  }

  initNTP(); // NOTE!!! call this BEFORE calling initTZsupport()
  // because configTime( ) has 0 for the hr and min offset

  initTZsupport(); // starts FS by calling initializeFS()
  output.flush();

#ifdef DEBUG
  Serial.print(" After initTZsupport ");
  printLocalSystemTime(&Serial);
#endif

  // normally no need to wait for NTP server as RTC sets time/date
  // just in case if not set wait here until set
  while (!isRTCDateTimeSet()) {
    delay(1);
    handleNTP(); // will sync system time as well if !isRTCDateTimeSet()
    output.flush();
  }
  // otherwise
  initCurrentFileName(); // set name for todays date

  Serial.print(" Time is set ");  // showTimeTZ(&Serial);
  Serial.print(" currentDateFileName: "); Serial.print(getCurrentDateFilename()); Serial.print("  ");
  printLocalSystemTime(&Serial);

  initProcessDateCsvFiles();

  processDataInit(); // loads history
  connectToHotWaterClient(); // first time skip reconnection delay
  output.flush();

  // Start the server
  server.begin();
#ifdef DEBUG
  Serial.println("Server started");
#endif

  // Print the IP address
#ifdef DEBUG
  Serial.println(WiFi.localIP());
#endif

#ifdef NO_TELNET_CONNECT
#ifdef DEBUG
  Serial.println(" Skip telnet connection to Hot Water Controller");
#endif
#endif

  startAsyncWebServer();

#ifdef DEBUG
  printLocalSystemTime(&Serial);
#endif
}

// not before dump or load the current data array is written so plenty of space for new data
// while processing dump of load
// called often from processing csv files  so add syncFromRTC and handleNTP here
void handleHotWaterRemoteClient() {
  syncFromRTC(); // keep system time in sync with RTC
  handleNTP();

  if (isRTCDateTimeSet()) { //skip connecting until we have correct time
    handleHotWaterClient();
  }
}

bool firstLoop = true;
millisDelay readTimeOut;
unsigned long READ_TIME_OUT_ms = 3000;

void loop() {
  //  .. other code that MUST run all the time
  output.nextByteOut(); // send bytes to Serial

  if (firstLoop) {
    firstLoop = false;
    loopTimer.clear(); // remove startup times
  }

  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }

  startTimers(0); // does nothing if started

  // Process CSV files in chunks to avoid blocking the main loop for too long
  processDateCsvFile();

  sw.update(); // call this every loop to update switch state

  if (sw.isChanged()) { // debounced switch changed state
    if (sw.isDown()) {
      unsigned long noRecs = appendBufferToFile(); // always
      (void)(noRecs); // to suppress unused variable warning
#ifdef DEBUG
      Serial.print("// Saved buffered records. ");  Serial.print(noRecs * 2); Serial.println(" secs of data.");
#endif
    }
  }

  handleHotWaterRemoteClient(); // call this often
  // handleHotWaterRemoteClient also calls
  // syncFromRTC(); // keep system time in sync with RTC
  // handleNTP();
  // to handle processing csv files


  if (sfCmdReader.read()) {
    if (!processCmd(sfCmdReader)) { // error dump rest of input
      Serial.println("// Flushing rest of input --- ");
      readTimeOut.start(READ_TIME_OUT_ms);
      sfCmdReader.setTimeout(2000);
      while (readTimeOut.isRunning()) {
        handleHotWaterRemoteClient(); // call often
        if (readTimeOut.justFinished()) {
          break;
        }
        if (sfCmdReader.read()) {
          readTimeOut.restart();
        }
      }
      Serial.println("// ===== end of input file ===== ");
    }
  }
  if (restartDelay.justFinished()) {
#ifdef DEBUG
    Serial.println("Rebooting...");
#endif
    hard_restart(); // includes call to appendBufferToFile();
  }

  if (hrCheckTimer.justFinished()) { // hrCheckTimer not started unless timeIsSet
    if (debugPtr) {
      debugPtr->print(" hrCheckTimer justFinished ");
    }
    // only if time has been set!!
    unsigned int minsNow = getLocalTime_mins();
    if (debugPtr) {
      debugPtr->print(" minsNow: "); debugPtr->println(minsNow);
    }
#ifdef DEBUG
    printLocalSystemTime(&Serial);
#endif
    // midnight within a min
    if ((minsNow == 0) || (minsNow == (24 * 60 - 1))) { // midnight within a min
      Serial.print(" rotated day history  ");
#ifdef DEBUG
      printLocalSystemTime(&Serial);
#endif
      // Write any remaining data to the file at midnight. NOTE: may also store first record of next day in previous day
      appendBufferToFile();
      rotateDayHistory();
      // clear last day's filename so next appendBufferToFile in about 1hr
      // will pick up new name
      clearCurrentDateFilename();
      restartDelay.start(RESTART_DELAY_ms); // restart in 2min from now
    }
    int excessSec = getLocalSecLeftInHr();
    if (excessSec > 5) { // allow to be 5sec early
      excessSec = 0;
    }
    timersStarted = false; // will restart now
    startTimers(excessSec); // restart
  }

  bool autoSetting = true;
  if (watts3PhKW_v >= 0.0) {
    autoSetting = false; // if importing then not auto
  }
  if (updated_v) {
    updated_v = false;
    _dwg_1.updateDwg_1(watts3PhKW_v, solarHotWaterKW_v, autoSetting);
  }
  handle_pfodServerConnection();  // may update state

  if (failedConnectRebootDelay.justFinished()) { // no connection for 1min
    esp_restart(); // force watch dog timer reboot
  }

}

bool validClient(WiFiClient & client) {
  return (client.connected());
}

void handle_pfodServerConnection() {
  if (server.hasClient()) { // new connection
#ifdef CLIENT_DEBUG
    Serial.print("new client:");
#endif
    bool foundSlot = false;
    size_t i = 0;
    for (; i < MAX_CLIENTS; i++) {
      if (!validClient(clients[i])) { // this space if free
        foundSlot = true;
        clients[i] = server.accept(); // was previously server.available();
        parserPtrs[i]->connect(bufferedClientPtrs[i]->connect(&(clients[i]))); // sets new io stream to read from and write to
        break;
      }
    }
    if (!foundSlot) {
      WiFiClient newClient = server.accept(); // was previously server.available(); // get any new client and close it
      newClient.stop();
#ifdef CLIENT_DEBUG
      Serial.println(" NO Slots available");
#endif
    } else {
#ifdef CLIENT_DEBUG
      Serial.println(i);
#endif
    }
  }
  for (size_t i = 0; i < MAX_CLIENTS; i++) {
    if (validClient(clients[i])) {
      handle_pfodParser(*parserPtrs[i]);
    }
  }
}

void closeConnection(Stream * io) {
  if (!io) {
#ifdef CLIENT_DEBUG
    Serial.println("closeConnection: Connection stream NULL");
#endif
    return;
  }
#ifdef CLIENT_DEBUG
  Serial.print("closeConnection:");
#endif
  bool foundSlot = false;
  size_t i = 0;
  for (; i < MAX_CLIENTS; i++) {
    if ((parserPtrs[i]->getPfodAppStream() == io) && validClient(clients[i])) {
      foundSlot = true;
      break;
    }
  }
  if (foundSlot) {
#ifdef CLIENT_DEBUG
    Serial.println(i);
#endif
    // found match
    parserPtrs[i]->closeConnection(); // nulls io stream
    bufferedClientPtrs[i]->stop(); // clears client reference
    clients[i].stop();
  } else {
#ifdef CLIENT_DEBUG
    Serial.println(" Connection stream NOT found");
#endif
  }
}

void setLable() {
  float localTotalCost = totalCost_v;
  float localYesterdayTotalCost = totalCost_yesterday_v;
  float localTotalHotWaterKWhrs = totalHotWaterKWhrs_v;
  float localYesterdayTotalHotWaterKWhrs = totalHotWaterKWhrs_yesterday_v;
  bool localWasHWfullyCharged = fullyHeated_v;
  bool localYesterdayWasHWfullyCharged = fullyHeated_yesterday_v;

  cSF(sfLabel, 50);
  if (state_v == 1) {
    sfLabel = "Solar Hot Water";
  } else if (state_v == 2) {
    sfLabel = "Off Peak Hot Water";
  } else if (state_v == 0) {
    sfLabel += "Hot Water OFF";
  } else {
    sfLabel += "No WiFi Connection to\nHot Water Controller";
  }
  sfLabel += "\n";
  dayMins_v = getLocalTime_mins();
  convertDayMinsToAmPm(sfTime, dayMins_v);
  sfLabel += sfTime;
  nameTimeDisplay.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Cost today ");
  if ((localTotalCost < 0.01) && (localTotalCost > -0.01)) {  // ~zero
    totalCost.setTextColor(0);
    localTotalCost = 0.0;
  } else if (localTotalCost > 0) {
    totalCost.setTextColor(9);
  } else {
    totalCost.setTextColor(2);
    sfLabel.print("-");
    localTotalCost = -localTotalCost;
  }
  sfLabel.print("$");
  sfLabel.print(localTotalCost, 2);
  totalCost.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Yesterday  ");
  if ((localYesterdayTotalCost < 0.01) && (localYesterdayTotalCost > -0.01)) {  // ~zero
    localYesterdayTotalCost = 0.0;
  } else if (localYesterdayTotalCost > 0) {
  } else {

    sfLabel.print("-");
    localYesterdayTotalCost = -localYesterdayTotalCost;
  }
  sfLabel.print("$");
  sfLabel.print(localYesterdayTotalCost, 2);

  sfLabel.print("  ");
  if (localYesterdayWasHWfullyCharged) {
    sfLabel.print("Fully Heated");
  } else {
    sfLabel.print(localYesterdayTotalHotWaterKWhrs, 1);
    sfLabel.print("kWhr");
  }
  yesterday.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Hot Water ");
  sfLabel.print(localTotalHotWaterKWhrs, 1);
  sfLabel.print("kWhr");
  hotwaterkWhrs.setText(sfLabel.c_str());
  if (localWasHWfullyCharged) {
    hotwaterFull.setText("Hot Water was Fully Heated");
  } else {
    hotwaterFull.setText("");
  }
}

void sendMainMenu(pfodParser & parser) {
  setLable();
  parser.print(F("{,"));  // start a Menu screen pfod message
  parser.print("<bg w>~"); // no prompt
  parser.sendRefreshAndVersion(dwgRefresh);
  // send menu items
  parser.print(F("|+A~"));
  parser.print(mainDwgLoadCmd); // the unique load cmd
  parser.print(F("}"));  // close pfod message
}

void sendMainMenuUpdate(pfodParser & parser) {
  setLable();
  parser.print(F("{;"));  // start an Update Menu pfod message
  parser.print(F("~")); // no change to colours and size
  // send menu items
  parser.print(F("|+A")); // reload A calls for dwg z update
  parser.print(F("}"));  // close pfod message
  // ============ end of menu ===========
}


int swap01(int in) {
  return (in == 0) ? 1 : 0;
}
// ============= end generated code =========

// telnet connection to SolarHotWater Controller IP 10.1.1.119 port 23

WiFiClient hotWaterClient; // for connection
millisDelay hotWaterReconnectDelay;
const unsigned long RECONNECT_DELAY_ms = 5000; // 5sec
millisDelay hotWaterClientKeepAliveTimer; //  to send keepalive msgs (empty lines)
const unsigned long KEEP_ALIVE_TIMER_ms = 5000; // 5sec
millisDelay connectionAliveDelay;
unsigned long CONNECTION_ALIVE_ms = 10000; // close connection is no msgs for 10sec

static int failedConnectionCount = 0;
static const int MAX_FAILED_CONNECTIONS = 5;

// returns true if connected else false
bool connectClientToHost(WiFiClient & client, const char* host, const uint16_t port) {
  appendBufferToFile(); // after a few of these esp32 E (84766) wifi:NAN WiFi stop
  if (!client.connect(host, port)) {
    if (debugPtr) {
      debugPtr->print("Connection to "); debugPtr->print(host); debugPtr->print(":"); debugPtr->print(port); debugPtr->println(" failed.");
    }
    if (!failedConnectRebootDelay.isRunning()) {
      failedConnectRebootDelay.start(FAILED_CONNECT_REBOOT_DELAY_MS); // restart in 1min if not connection
    }
    state_v = 3; // no wifi connection
    failedConnectionCount++;
    if (failedConnectionCount > MAX_FAILED_CONNECTIONS) {
      hard_restart();
    }
    return false;
  }
  if (debugPtr) {
    debugPtr->print("Connected to "); debugPtr->print(host); debugPtr->print(":"); debugPtr->print(port); debugPtr->println();
  }
  failedConnectionCount = 0;
  failedConnectRebootDelay.stop();
  return true;
}

bool justConnected = false; // skip first two line of welcome message

createSafeStringReader(sfReaderHotWater, 120, "\n\r");

// skip if #define NO_TELNET_CONNECT
void connectToHotWaterClient() {
#ifndef NO_TELNET_CONNECT
  if (connectClientToHost(hotWaterClient, HotWater_host, HotWater_portNo)) {
    hotWaterClient.println();
    hotWaterClient.setTimeout(5);
    hotWaterClientKeepAliveTimer.start(KEEP_ALIVE_TIMER_ms);
    sfReaderHotWater.connect(hotWaterClient);
    //  sfReaderHotWater.returnEmptyTokens(); // empty lines keep connection open
    connectionAliveDelay.start(CONNECTION_ALIVE_ms);
  } else {
    hotWaterClient.stop();
    return;
  }
#endif
}

// comment out handling if if  #define NO_TELNET_CONNECT
static void handleHotWaterClient() {
#ifdef  HOT_WATER_CLIENT_TIME_CHECK
  loopTimer.check(Serial);
#endif
#ifndef NO_TELNET_CONNECT
  if (!hotWaterClient.connected()) {
    connectionAliveDelay.stop();
    if (hotWaterClientKeepAliveTimer.isRunning()) {
      hotWaterClientKeepAliveTimer.stop();
      sfReaderHotWater.end();
      hotWaterClient.stop();
    }
    if (!hotWaterReconnectDelay.isRunning()) {
      hotWaterReconnectDelay.start(RECONNECT_DELAY_ms);
      if (debugPtr) {
        debugPtr->print("start reconnection delay"); debugPtr->println();
      }
    }
  }
  if (hotWaterReconnectDelay.justFinished()) {
    if (debugPtr) {
      debugPtr->print("reconnection delay finished"); debugPtr->println();
    }
    connectToHotWaterClient();
    loopTimer.clear(); // remove startup times
  }
  if (hotWaterClient.connected()) {
    if (sfReaderHotWater.read()) {
      connectionAliveDelay.restart();
      sfReaderHotWater.trim();
      processData(sfReaderHotWater);
    } else { // did not find a line of input
      if (connectionAliveDelay.justFinished()) {
        hotWaterClient.stop(); // stop and try to reconnect
        sfReaderHotWater.end();
      }
    }
    if (hotWaterClientKeepAliveTimer.justFinished()) {
      hotWaterClientKeepAliveTimer.restart();
      hotWaterClient.println();
    }
  }
#endif
}
