#include "csvProcessor.h"
#include "RTC_NTP_replacement.h"
#include "HistoryFile.h"
#include "HW_webpages.h"
#include <vector>


//#define DEBUG

// Global variables for CSV file processing
static std::vector<String> csvFiles;
static size_t currentCsvFileIndex = 0;
static bool csvProcessingInitialized = false;
static bool finished = false; // suppress end msgs
static Print* debugPtr = NULL;

// Look for /YYYYMMDD.csv files to merge into /YYYYMMDD
// NOTE the resulting /YYYYMMDD file is the one specified in the //FileName: line in the csv
// NOT the name of the csv file !!
void initProcessDateCsvFiles() {
#ifdef DEBUG
  debugPtr = getDebugOut();
#endif

  csvFiles.clear();
  currentCsvFileIndex = 0;
  csvProcessingInitialized = true;
  finished = false;

  // List all files in the root directory and filter for date format
  File root = LittleFS.open("/");
  if (!root || !root.isDirectory()) {
    Serial.println("Failed to open root dir");
    return;
  }

  File file = root.openNextFile();
  while (file) {
    String fileName = String(file.name());
    if (isFileNameDateCsv(fileName.c_str())) {
      // add / to front
      String filePath = '/' + fileName;
      csvFiles.push_back(filePath);
    }
    file = root.openNextFile();
  }
  root.close();

  // If we found no matching date files, return empty data
  if (csvFiles.empty()) {
    Serial.println("No date.csv files to process");
    return;
  }

  Serial.print("Found "); Serial.print(csvFiles.size()); Serial.println(" date.csv files to process");
}

// Process one date CSV file at a time
// Returns true if there are more files to process
bool processDateCsvFile() {
  File currentDateFile;

  if (finished) {
    return false;
  }
  // If not initialized or no files remaining, return false
  if (!csvProcessingInitialized) {
    Serial.println(" Need to call initProcessDateCsvFiles() first.");
    return false;
  }

  if (currentCsvFileIndex >= csvFiles.size()) {
    Serial.println(" Completed process all date .csv files");
    csvProcessingInitialized = false;
    finished = true;
    return false;
  }
  
  appendBufferToFile(); // incase merging to existing file do this before opening files

  // Process the next file in the list
  const String& filePath = csvFiles[currentCsvFileIndex];
  currentDateFile = LittleFS.open(filePath, FILE_READ);
  if (!currentDateFile) {
    Serial.printf("Failed to open %s for reading\r\n", filePath.c_str());
    currentCsvFileIndex++; // Move to next file
    return (currentCsvFileIndex < csvFiles.size()); // Return if more files to process
  }

  Serial.print("Processing "); Serial.println(filePath);

  // Process the entire file at once
  bool result = readDataStream(currentDateFile, &Serial); // prints error msgs
  if (!result) {
    if (debugPtr) {
      debugPtr->print(" readDataStream returned false for: "); debugPtr->println(filePath);
    }
  }
  // Clean up after processing
  currentDateFile.flush(); // need to flush here or delete does not work (why??)
  currentDateFile.close();

  // delete the file
  if (LittleFS.remove(filePath)) {
    Serial.print("Deleted "); Serial.print(filePath); Serial.println(" after processing");
  } else {
    Serial.print("Error deleting "); Serial.println(filePath);
  }

  // Move to the next file
  currentCsvFileIndex++;
  return true; // Return if more files to process
}
